.discsongsflowbox > flowboxchild {
    padding: 0px;
}

/* AlbumCover */
.albumcover-artist-label {
    font-size: smaller;
}

/* By default in Adwaita dark this is not transparent */
.tile {
    background-color: transparent;
}

/* AlbumWidget */
.title-artist {
    font-size: larger;
    font-weight: bold;
}

.disc-label {
    background-color: @theme_bg_color;
    color: alpha(@theme_fg_color, 0.8);
    padding: 12px 0;
}

/* ArtistAlbumsWidget */
box#ArtistAlbumsWidget .artist-label {
    font-weight: bold;
}

/* NoMusic */
.no-music-found-label {
    font-size: xx-large;
    font-weight: bold;
}

/* PlayerToolbar */
.player-title-label {
    font-weight: bold;
    font-size: smaller;
}

.player-artist-label {
    font-size: smaller;
}

/* PlaylistControls */
.playlist-name-label {
  font-weight: bold;
  font-size: x-large;
}

/* ArtistAlbumWidget */
.artist-albums-widget,
.artist-albums-widget:hover {
    background-color: @theme_bg_color;
}

.artist-albums-widget > row:hover {
    background-color: rgba(0, 0, 0, 0.0);
    box-shadow: none;
}

.album-title {
    font-size: large;
    font-weight: bold;
}

/* FIXME: Remove once songsview is ported to the new style */
.songs-list-old {
    box-shadow: inset 0 -1px shade(@borders, 1.30);
    background-color: @theme_bg_color;
}

.songs-list-old:selected {
    color: @theme_fg_color;
    border-color: mix(@theme_fg_color, @theme_bg_color, 0.5);
}

/* We use background-image as a workaround on the StarImage widget to
   enable transitions between the non-starred and starred icons. */
.star {
    background-image: -gtk-icontheme('non-starred-symbolic');
    color: alpha(@theme_fg_color, 0.3);
    transition: 100ms linear;
}

.star:hover {
    color: alpha(@theme_fg_color, 0.6);
}

.star:selected {
    background-image: -gtk-icontheme('starred-symbolic');
    color: @theme_selected_bg_color;
    transition: 100ms linear;
}

.content-view { background-color: @theme_bg; }

.grey-image {
    color: alpha(@theme_fg_color, 0.3);
}

/* Used in AlbumWidget and ArtistAlbumsWidget */
.playing-song-label {
    font-weight: bold;
}

/* TwoLineTip */
.tooltip-title {
  font-weight: bold;
}

/* Lists style */

/* workaround to avoid a black background issue
in AlbumWidget and PlaylistsView
https://gitlab.gnome.org/GNOME/gtk/issues/694
*/
list {
    background-color: transparent;
}

/* workaround to avoid a black background issue
in AlbumWidget and PlaylistsView
https://gitlab.gnome.org/GNOME/gtk/issues/694 */
.songs-list > row {
    background-color: @theme_base_color;
}

.songs-list > row:selected label {
    color: @theme_fg_color;
}


.disc-list-box > row {
padding: 0px;
}

.songs-list {
    border: 1px solid rgba(0, 0, 0, 0.1);
}

.songs-list > row {
    border-bottom: 1px solid rgba(0, 0, 0, 0.1);
}

.songs-list > row:last-child {
    border-bottom: none;
}

.playlistdialog-row:selected {
    color: @theme_fg_color;
    background-color: @theme_insensitive_bg_color;
}

.playlistdialog-row:selected label {
    color: @theme_text_color;
}

/* SearchView */
.search-header {
  font-weight: bold;
}
